#include "stdafx.h"
#using <mscorlib.dll>
#using <System.dll>
#using <System.Drawing.DLL>
#using <System.Windows.Forms.DLL>
#include <windows.h>
#include <ShlObj.h>
#include "SendToMenuItem.h"

using namespace System;
using namespace System::Collections;
using namespace System::Drawing;
using namespace System::Windows::Forms;
using namespace System::Runtime::InteropServices;

namespace megaBYTE {
  private __gc class CSendToMenuItem {
  public:
    LPCITEMIDLIST m_pidl;
    Bitmap *m_pbmpNormal;
  };

  public __gc class SendToMenuItem;
  
  __nogc class CSendToMenu {
  protected:
    IShellFolder *m_psfDesktop;
    IMalloc *m_pMalloc;
  public:
    CSendToMenu(public __gc class SendToMenuItem *pParent);
    ~CSendToMenu();
  protected:
    inline void ShellFree(void *pv);
    LPSHELLFOLDER GetSendToFolder(void);
    void FillMenu(public __gc class SendToMenuItem *pParent);
    void StrRetToStr(LPSTRRET pstr, LPITEMIDLIST pidl);
    inline HRESULT GetUIObjectOfPath(LPOLESTR pFileName, REFIID riid, LPVOID *ppvOut);
    inline LPITEMIDLIST PidlFromPath(LPOLESTR pFileName);
    void DoDrop(LPDATAOBJECT pdto, LPDROPTARGET pdt);
    inline LPITEMIDLIST StepOverPidl(LPITEMIDLIST pidl);
    HRESULT GetUIObjectOfAbsPidl(LPITEMIDLIST pidl, REFIID riid, LPVOID *ppvOut);
  public:
    void DoItemSelect(LPOLESTR pFileName, LPCITEMIDLIST pidl);
  };
  
  public __gc class SendToClickEventArgs : public System::EventArgs {
  protected:
    String *m_strFileName;
  public:
  __property String *get_FileName(void) { return m_strFileName; }
  __property void set_FileName(String *str) { m_strFileName = str; }
  };

  public __gc class SendToMenuItem : public MenuItem {
  protected:
    const int m_iHorGap;
    ArrayList *m_pItems;
    CSendToMenu *m_pWrap;
    bool m_bEnableBitmap;
    MenuItem *m_pReplacedItem;
    EventHandler *m_pReplacedClickEventHandler;
  public:
    SendToMenuItem();
  protected:
    ~SendToMenuItem();
  private public:
    void AddItem(String *strName, Bitmap *pBmp, LPCITEMIDLIST pidl);
  protected:
    void CustomMeasureItem(Object *sender, MeasureItemEventArgs *e);
    void CustomDrawItem(Object *sender, DrawItemEventArgs *e);
    void CustomClickItem(Object *sender, EventArgs *e);
  public:
    __property bool get_EnableBitmap(void) { return m_bEnableBitmap; }
    __property void set_EnableBitmap(bool bEnable) {
      if (bEnable != m_bEnableBitmap) {
        m_bEnableBitmap = bEnable;
        for (int i = 0; i < MenuItems->Count; i++) {
          MenuItems->Item[i]->OwnerDraw = m_bEnableBitmap;
        }
      }
    }
    void Replace(MenuItem *menuItem, EventHandler *eventHandler);
    void Restore(void);
  };
}

namespace megaBYTE {
  CSendToMenu::CSendToMenu(public __gc class SendToMenuItem *pParent) {
    if (FAILED(::SHGetMalloc(&m_pMalloc))) {
      m_pMalloc = NULL;
    }
    if (FAILED(::SHGetDesktopFolder(&m_psfDesktop))) {
      m_psfDesktop = NULL;
    }
    FillMenu(pParent);
  }
  
  CSendToMenu::~CSendToMenu() {
    if (m_psfDesktop != NULL) {
      m_psfDesktop->Release();
    }
    if (m_pMalloc != NULL) {
      m_pMalloc->Release();
    }
  }
  
  inline void CSendToMenu::ShellFree(void *pv) {
    if (m_pMalloc != NULL) {
      m_pMalloc->Free(pv);
    }
  }
  
  LPSHELLFOLDER CSendToMenu::GetSendToFolder(void) {
    LPSHELLFOLDER psf = NULL;
    LPITEMIDLIST pidl = NULL;
    HRESULT hres = ::SHGetSpecialFolderLocation(NULL, CSIDL_SENDTO, &pidl);
    if (SUCCEEDED(hres)) {
      if (FAILED(m_psfDesktop->BindToObject(pidl, NULL, IID_IShellFolder, (LPVOID *)&psf))) {
        psf = NULL;
      }
      ShellFree(pidl);
    }
    return psf;
  }
  
  void CSendToMenu::StrRetToStr(LPSTRRET pstr, LPITEMIDLIST pidl) {
    LPWSTR pOleStr;
    switch (pstr->uType) {
    case STRRET_WSTR:
      /*
      *      The pOleStr points to a WCHAR string that we convert
      *      to ANSI.  We need to stash pstr->pOleStr into a safe
      *      place, because pstr->cStr lives in the same place.
      */
      pOleStr = pstr->pOleStr;
      WideCharToMultiByte(CP_ACP, 0, pOleStr, -1, pstr->cStr, MAX_PATH, NULL, NULL);
      ShellFree(pOleStr);
      break;
    case STRRET_OFFSET:
      /*
      *      The string lives inside the pidl, so copy it out.
      */
      lstrcpyn(pstr->cStr, (LPSTR)((LPBYTE)pidl + pstr->uOffset), MAX_PATH);
      break;
    case STRRET_CSTR:
      /*
      *      The string already is in the right place.
      */
      break;
    }
  }

  void CSendToMenu::FillMenu(public __gc class SendToMenuItem *pParent) {
    LPSHELLFOLDER psf = GetSendToFolder();
    if (psf != NULL) {
      LPENUMIDLIST peidl;
      if (SUCCEEDED(psf->EnumObjects(NULL, SHCONTF_FOLDERS | SHCONTF_NONFOLDERS, &peidl))) {
        LPITEMIDLIST pidl;
        while (peidl->Next(1, &pidl, NULL) == S_OK) {
          STRRET str;
          SHFILEINFO shi;
          if (SUCCEEDED(psf->GetDisplayNameOf(pidl, SHGDN_NORMAL, &str)) &&
              SHGetFileInfo((LPCSTR)pidl, 0, &shi, sizeof(shi), SHGFI_ICON |
                            SHGFI_SHELLICONSIZE | SHGFI_SMALLICON |SHGFI_PIDL) != 0) {
            StrRetToStr(&str, pidl);
            String *pStr = new String(str.cStr);
            Bitmap *pBmp = Bitmap::FromHicon(shi.hIcon);
            pParent->AddItem(pStr, pBmp, pidl);
          }
        }
        peidl->Release();
      }
      psf->Release();
    }
  }

  inline LPITEMIDLIST CSendToMenu::PidlFromPath(LPOLESTR pFileName) {
    LPITEMIDLIST pidl;
    ULONG ulEaten;
    DWORD dwAttributes;
    if (FAILED(m_psfDesktop->ParseDisplayName(NULL, NULL, pFileName, &ulEaten, &pidl,
                                              &dwAttributes))) {
      return NULL;
    }
    return pidl;
  }

  inline LPITEMIDLIST CSendToMenu::StepOverPidl(LPITEMIDLIST pidl) {
    return (LPITEMIDLIST)((BYTE*)pidl + pidl->mkid.cb);
  }

  HRESULT CSendToMenu::GetUIObjectOfAbsPidl(LPITEMIDLIST pidl, REFIID riid, LPVOID *ppvOut) {
    *ppvOut = NULL;
    LPITEMIDLIST pidlLast, pidlNext;
    for (pidlLast = pidl; pidlNext = StepOverPidl(pidlLast), pidlNext->mkid.cb;
        pidlLast = pidlNext) {
    }
    USHORT cb = pidlLast->mkid.cb;
    pidlLast->mkid.cb = 0;
    HRESULT hres = S_OK;
    LPSHELLFOLDER psf;
    if (pidl == pidlLast) {
      psf = m_psfDesktop;
      m_psfDesktop->AddRef();
    }
    else {
      hres = m_psfDesktop->BindToObject(pidl, NULL, IID_IShellFolder, (LPVOID *)&psf);
    }
    pidlLast->mkid.cb = cb;
    if (FAILED(hres)) {
      return hres;
    }
    hres = psf->GetUIObjectOf(NULL, 1, (LPCITEMIDLIST *)&pidlLast, riid, NULL, ppvOut);
    psf->Release();
    return hres;
  }

  inline HRESULT CSendToMenu::GetUIObjectOfPath(LPOLESTR pFileName, REFIID riid,
                                                LPVOID *ppvOut) {
    *ppvOut = NULL;
    LPITEMIDLIST pidl = PidlFromPath(pFileName);
    if (pidl == NULL) {
      return E_FAIL;
    }
    HRESULT hres = GetUIObjectOfAbsPidl(pidl, riid, ppvOut);
    ShellFree(pidl);
    return hres;
  }

  void CSendToMenu::DoDrop(LPDATAOBJECT pdto, LPDROPTARGET pdt) {
    POINTL pt = { 0, 0 };
    DWORD dwEffect = DROPEFFECT_COPY | DROPEFFECT_MOVE | DROPEFFECT_LINK;
    if (SUCCEEDED(pdt->DragEnter(pdto, MK_LBUTTON, pt, &dwEffect)) && dwEffect) {
      pdt->Drop(pdto, MK_LBUTTON, pt, &dwEffect);
    }
    else {
      pdt->DragLeave();
    }
  }

  void CSendToMenu::DoItemSelect(LPOLESTR pFileName, LPCITEMIDLIST pidl) {
    LPDATAOBJECT pdto;
    if (SUCCEEDED(GetUIObjectOfPath(pFileName, IID_IDataObject, (LPVOID *)&pdto))) {
      LPSHELLFOLDER psf = GetSendToFolder();
      if (psf != NULL) {
        LPDROPTARGET pdt;
        if (SUCCEEDED(psf->GetUIObjectOf(NULL, 1, &pidl, IID_IDropTarget, NULL,
                                          (LPVOID *)&pdt))) {
          DoDrop(pdto, pdt);
          pdt->Release();
        }
        psf->Release();
      }
      pdto->Release();
    }
  }

  ///////////////////////////////////////////////////////////////////////////////////////

  SendToMenuItem::SendToMenuItem() : MenuItem(), m_bEnableBitmap(true), m_iHorGap(4),
                                    m_pReplacedItem(NULL), m_pReplacedClickEventHandler(NULL) {
    m_pItems = new ArrayList();
    m_pWrap = new CSendToMenu(this);
  }

  SendToMenuItem::~SendToMenuItem() {
    Restore();
    delete m_pWrap;
  }

  void SendToMenuItem::AddItem(String *strName, Bitmap *pBmp, LPCITEMIDLIST pidl) {
    CSendToMenuItem *ptr = new CSendToMenuItem();
    ptr->m_pbmpNormal = pBmp;
    ptr->m_pidl = pidl;
    m_pItems->Add(ptr);
    MenuItem *pmnuItem = new MenuItem(strName);
    pmnuItem->OwnerDraw = m_bEnableBitmap;
    pmnuItem->DrawItem += new DrawItemEventHandler(this, CustomDrawItem);
    pmnuItem->MeasureItem += new MeasureItemEventHandler(this, CustomMeasureItem);
    pmnuItem->Click += new EventHandler(this, CustomClickItem);
    MenuItems->Add(pmnuItem);
  }

  void SendToMenuItem::CustomMeasureItem(Object *sender, MeasureItemEventArgs *e) {
    SizeF szFont = e->Graphics->MeasureString(MenuItems->Item[e->Index]->Text, SystemInformation::MenuFont);
    Size sz = szFont.ToSize();
    CSendToMenuItem *ptr = static_cast<CSendToMenuItem *>(m_pItems->Item[e->Index]);
    e->ItemHeight = System::Math::Max(sz.Height, ptr->m_pbmpNormal->Height);
    e->ItemHeight = System::Math::Max(e->ItemHeight, SystemInformation::MenuHeight);
    e->ItemWidth = sz.Width + ptr->m_pbmpNormal->Width + m_iHorGap * 2;
  }

  void SendToMenuItem::CustomDrawItem(Object *sender, DrawItemEventArgs *e) {
    e->DrawBackground();
    if (e->State & DrawItemState::Focus) {
      e->DrawFocusRectangle();
    }
    Color crMenuText;
    if (e->State & DrawItemState::Disabled) {
      crMenuText = Color(SystemColors::GrayText);
    }
    else if (e->State & DrawItemState::Disabled) {
      crMenuText = Color(SystemColors::HighlightText);
    }
    else {
      crMenuText = Color(SystemColors::MenuText);
    }
    SolidBrush *brMenuText = new SolidBrush(crMenuText);
    CSendToMenuItem *ptr = static_cast<CSendToMenuItem *>(m_pItems->Item[e->Index]);
    Point pnt(e->Bounds.X + m_iHorGap, e->Bounds.Y + (e->Bounds.Height - ptr->m_pbmpNormal->Height) / 2 + 1);
    e->Graphics->DrawImageUnscaled(ptr->m_pbmpNormal, pnt);
    String &str = *MenuItems->Item[e->Index]->Text;
    SizeF szFont = e->Graphics->MeasureString(&str, SystemInformation::MenuFont);
    Size sz = szFont.ToSize();
    RectangleF rect(e->Bounds.X + ptr->m_pbmpNormal->Width + m_iHorGap * 2, 
                    e->Bounds.Y + (e->Bounds.Height - sz.Height) / 2 + 1,
                    e->Bounds.Width, e->Bounds.Height);
    e->Graphics->DrawString(&str, e->Font, brMenuText, rect);
  }

  void SendToMenuItem::CustomClickItem(Object *sender, EventArgs *e) {
    if (MenuItem *mi = dynamic_cast<MenuItem *>(sender)) {
      CSendToMenuItem *ptr = static_cast<CSendToMenuItem *>(m_pItems->Item[mi->Index]);
      SendToClickEventArgs *newE = new SendToClickEventArgs();
      OnClick(newE);
      LPOLESTR pFileName = static_cast<LPOLESTR>(Marshal::StringToCoTaskMemUni(newE->FileName).ToPointer());
      m_pWrap->DoItemSelect(pFileName, ptr->m_pidl);
      Marshal::FreeCoTaskMem(pFileName);
    }
  };

  void SendToMenuItem::Replace(MenuItem *menuItem, EventHandler *eventHandler) {
    if (menuItem == NULL || eventHandler == NULL) {
      throw new System::ArgumentNullException();
    }
    Restore();
    menuItem->Parent->MenuItems->Add(this);
    BarBreak = menuItem->BarBreak;
    Break = menuItem->Break;
    DefaultItem = menuItem->DefaultItem;
    Enabled = menuItem->Enabled;
    Index = menuItem->Index;
    MdiList = menuItem->MdiList;
    MergeOrder = menuItem->MergeOrder;
    MergeType = menuItem->MergeType;
    Shortcut = menuItem->Shortcut;
    ShowShortcut = menuItem->ShowShortcut;
    Text = menuItem->Text;
    Visible = menuItem->Visible;
    m_pReplacedClickEventHandler = eventHandler;
    Click += eventHandler;
    m_pReplacedItem = menuItem;
    menuItem->Parent->MenuItems->Remove(menuItem);
  }

  void SendToMenuItem::Restore(void) {
    if (m_pReplacedItem != NULL) {
      Click -= m_pReplacedClickEventHandler;
      m_pReplacedClickEventHandler = NULL;
      Parent->MenuItems->Add(m_pReplacedItem);
      m_pReplacedItem->Index = Index;
      Parent->MenuItems->Remove(this);
      m_pReplacedItem = NULL;
    }
  }
}
